import json
import pytest
from transmute_core.handler import process_result
from transmute_core import default_context, APIException

CONTENT_TYPE = "application/json"


class SomeOtherException(Exception):
    pass


def test_process_result_200(complex_transmute_func):
    result = {"kind": "dog", "age": 5}
    exc = None
    output = process_result(
        complex_transmute_func,
        default_context, result,
        exc, CONTENT_TYPE
    )
    assert json.loads(output["body"].decode()) == result
    assert output["code"] == 200
    assert output["content-type"] == CONTENT_TYPE


def test_process_result_api_exception(complex_transmute_func):
    result = {"kind": "dog", "age": 5}
    exc = APIException("foo")
    output = process_result(
        complex_transmute_func,
        default_context, result,
        exc, CONTENT_TYPE
    )
    assert json.loads(output["body"].decode()) == {
        "result": "invalid api use: " + str(exc),
        "success": False,
        "code": 400
    }
    assert output["code"] == 400
    assert output["content-type"] == CONTENT_TYPE


def test_process_general_exception(complex_transmute_func):
    result = {"kind": "dog", "age": 5}
    exc = SomeOtherException("foo")
    with pytest.raises(SomeOtherException):
        process_result(
            complex_transmute_func,
            default_context, result,
            exc, CONTENT_TYPE
        )


def test_unknown_content_type_defaults_to_json(complex_transmute_func):
    result = {"kind": "dog", "age": 5}
    exc = None
    output = process_result(
        complex_transmute_func,
        default_context, result,
        exc, "application/myson"
    )
    assert json.loads(output["body"].decode()) == result
    assert output["content-type"] == "application/json"



import pytest
from transmute_core.function import TransmuteFunction
import transmute_core


@transmute_core.describe(paths="/")
@transmute_core.annotate({"return": int})
def raw_func():
    return 12345


@pytest.fixture
def func():
    return TransmuteFunction(raw_func)


def test_callable(func):
    """ test function is callable, and routes to the inner function. """
    assert func() == 12345


def test_return_type(func):
    """ test function is callable, and routes to the inner function. """
    assert func.return_type is int


def test_swagger_schema_has_object(func):
    swagger = func.get_swagger_operation()
    assert swagger.responses["200"].schema.to_primitive() == {
        "type": "number"
    }


def test_swagger_schema_path(func):
    swagger = func.get_swagger_path()
    assert swagger.get.responses["200"].schema.to_primitive() == {
        "type": "number"
    }


def test_function_raises_exception_on_path_missing():

    @transmute_core.describe(paths=[])
    def func():
        pass

    with pytest.raises(transmute_core.InvalidTransmuteDefinition):
        TransmuteFunction(func)



import pytest
from transmute_core import (
    annotate, describe,
    default_context,
    SchematicsSerializer,
    get_default_serializer_set,
    ResponseShapeComplex,
    TransmuteFunction
)
from schematics.models import Model
from schematics.types import StringType, IntType


@pytest.fixture
def serializer_set():
    return get_default_serializer_set()


@pytest.fixture
def serializer():
    return SchematicsSerializer()


@pytest.fixture
def transmute_func():

    @describe(paths="/api/v1/multiply")
    @annotate({"left": int, "right": int, "return": int})
    def multiply(left, right):
        return left * right

    return TransmuteFunction(multiply)


class Pet(Model):

    kind = StringType(required=True)
    age = IntType()


@pytest.fixture
def complex_transmute_func():

    @describe(paths="/api/v1/adopt")
    @annotate({"return": Pet})
    def adopt():
        return Pet({
            "kind": "dog",
            "age": 5
        })

    return TransmuteFunction(adopt)



from transmute_core.function.attributes import TransmuteAttributes


def test_merge():
    left = TransmuteAttributes(
        methods=["POST"], query_parameters=["a"]
    )
    right = TransmuteAttributes(
        methods=["PUT"], query_parameters=["b"],
        body_parameters=["c"]
    )
    joined = left | right
    assert joined.methods == set(["POST", "PUT"])
    assert joined.query_parameters == set(["a", "b"])
    assert joined.body_parameters == set(["c"])



from transmute_core.swagger import (
    generate_swagger_html, get_swagger_static_root,
    SwaggerSpec
)
from transmute_core import default_context


def test_generate_swagger():
    swagger_static_root = "/statics/_swagger"
    swagger_json_url = "/swagger.json"
    body = generate_swagger_html(swagger_static_root, swagger_json_url)
    assert swagger_static_root in body
    assert swagger_json_url in body


def test_get_swagger_static_root():
    assert "static" in get_swagger_static_root()


def test_swagger_definition_generation():
    """
    swagger routes should be ablo to generate a proper
    spec.
    """
    routes = SwaggerSpec()
    assert routes.swagger_definition() == {
        "info": {"title": "example", "version": "1.0"},
        "paths": {},
        "swagger": "2.0",
        "basePath": "/"
    }


def test_swagger_transmute_func(transmute_func):
    """
    swagger routes should be ablo to generate a proper
    spec.
    """
    routes = SwaggerSpec()
    routes.add_func(transmute_func, default_context)
    assert routes.swagger_definition() == {
        "info": {"title": "example", "version": "1.0"},
        "paths": {
            "/api/v1/multiply": transmute_func.get_swagger_path(default_context).to_primitive(),
        },
        "swagger": "2.0",
        "basePath": "/"
    }



class WebTransmuteException(Exception):
    """ base exception """


class APIException(WebTransmuteException):
    """
    this exception signifies an exception
    with API
    """

    def __init__(self, message, code=400):
        """ code can be overriden, to allow the proper status code. """
        super(APIException, self).__init__(message)
        self.code = code


class SerializationException(APIException):
    """
    this exceptions signifies an exception with
    serializing values.
    """


class NoSerializerFound(WebTransmuteException):
    """
    raised when a serializer does not exist to handle the
    desired content type.
    """


class InvalidTransmuteDefinition(WebTransmuteException):
    """
    this exception is raised when an invalid configuration for
    transmute is encountered
    """



class ObjectSerializer(object):
    """
    The object serializer is responsible for converting objects to and
    from basic data types. Basic data types are serializable to and
    from most common data representation languages (such as yaml or json)

    Basic data types are:

    - str (basestring in Python2, str in Python3)
    - float
    - int
    - None
    - dict
    - list

    The serializer decides what it can and can not serialize, and should raise
    an exception when a type it can not serialize is passed.

    `SchematicsSerializer` is the default implementation used.
    """

    def load(self, model, value):
        """
        load the value from a basic datatype, into a class.

        if the model or value is not valid, raise a SerializationException
        """
        raise NotImplementedError()

    def dump(self, model, value):
        """
        dump the value from a class to a basic datatype.

        if the model or value is not valid, raise a SerializationException
        """
        raise NotImplementedError()

    def to_json_schema(self, model):
        """
        return a dictionary representing a jsonschema for the model.
        """
        raise NotImplementedError()



"""
utilities to help with generating handlers.
"""
from .exceptions import (
    APIException, NoSerializerFound
)


def process_result(transmute_func, context, result, exc, content_type):
    """ process a result """
    output = None
    code = 200
    if exc:
        if isinstance(exc, APIException):
            output = {
                "result": "invalid api use: {0}".format(str(exc)),
                "success": False,
            }
            code = exc.code
        else:
            # TODO: better exception raising,
            # to keep traceback context.
            raise exc
    else:
        if transmute_func.return_type:
            result = context.serializers.dump(
                transmute_func.return_type, result
            )
        output = {
            "result": result,
            "success": True
        }
    try:
        serializer = context.contenttype_serializers[content_type]
    except NoSerializerFound:
        serializer = context.contenttype_serializers.default
        content_type = serializer.main_type
    output["code"] = code
    if output["success"]:
        result = context.response_shape.create_body(output)
    else:
        result = output
    body = serializer.dump(result)
    # keeping the return type a dict to
    # reduce performance overhead.
    return {
        "body": body,
        "code": code,
        "content-type": content_type
    }



import pytest

from transmute_core.decorators import (
    describe, annotate
)


@pytest.mark.parametrize("method", [
    "PUT", "POST", "DELETE"
])
def test_describe(method):

    @describe(methods=method)
    def test():
        pass

    assert test.transmute.methods == set([method])


def test_describe_join():
    """ two describe annotations should merge the descriptions """
    @describe(methods=["GET"])
    @describe(methods=["PUT"])
    def test():
        pass

    assert test.transmute.methods == set(["GET", "PUT"])


def test_annotate():

    annotations = {"return": str, "arg": int}

    @annotate(annotations)
    def delete_test(arg):
        pass

    assert delete_test.__annotations__ == annotations



import pytest
from schematics.models import Model
from schematics.types import StringType, IntType
from schematics.types.compound import DictType
from schematics.exceptions import ValidationError
from transmute_core.exceptions import SerializationException


class Card(Model):
    name = StringType()
    price = IntType()


def greater_than_zero(value):
    if value < 0:
        raise ValidationError("must be greater than zero")
    return value


class Person(Model):
    age = IntType(required=True, validators=[greater_than_zero])
    bio = StringType()

card_dict = {"name": "foo", "price": 10}
card = Card(card_dict)


@pytest.mark.parametrize("typ,inp,out", [
    (Card, card, card_dict),
    ([Card], [card], [card_dict])
])
def test_schematics_integration_dump(serializer, typ, inp, out):
    assert serializer.dump(typ, inp) == out


def test_schematics_validate_is_called(serializer):
    with pytest.raises(SerializationException):
        serializer.load(Person, {"age": -1, "bio": "foo"})


@pytest.mark.parametrize("typ,inp,out", [
    (Card, card_dict, card),
    ([Card], [card_dict], [card])
])
def test_schematics_integration_load(serializer, typ, inp, out):
    assert serializer.load(typ, inp) == out


def test_schematics_to_json_schema(serializer):
    assert serializer.to_json_schema(Card) == {
        "properties": {
            "name": {"type": "string"},
            "price": {"type": "number"}
        },
        "title": "Card",
        "type": "object"
    }


def test_schematics_to_json_schema_required_value(serializer):

    class CardNameRequired(Model):
        name = StringType(required=True)
        price = IntType()

    assert serializer.to_json_schema(CardNameRequired) == {
        "properties": {
            "name": {"type": "string"},
            "price": {"type": "number"}
        },
        "title": "CardNameRequired",
        "required": ["name"],
        "type": "object"
    }


def test_to_json_list(serializer):
    assert serializer.to_json_schema([int]) == {
        "type": "array",
        "items": {"type": "number"}
    }


def test_to_json_dict(serializer):
    assert serializer.to_json_schema(DictType(IntType())) == {
        "type": "object",
        "additionalProperties": {"type": "number"}
    }


def test_serialize_type(serializer):
    """ serializer should be able to serialize a raw type. """
    assert serializer.dump(DictType(StringType()), {"key": "foo"}) == {"key": "foo"}



from .interface import ObjectSerializer
from collections import OrderedDict
from schematics.types import (
    BaseType,
    BooleanType,
    DecimalType,
    FloatType,
    IntType,
    NumberType,
    StringType
)
from schematics.models import Model, ModelMeta
from schematics.types.compound import (
    CompoundType, ListType, ModelType, DictType
)
from schematics.exceptions import BaseError
from schematics.transforms import get_import_context
from ..exceptions import SerializationException
from decimal import Decimal
from ..compat import all_string_types

MODEL_MAP = {
    int: IntType(),
    bool: BooleanType(),
    float: FloatType(),
    Decimal: DecimalType(),
    None: BaseType()
}

for t in all_string_types:
    MODEL_MAP[t] = StringType()


JSON_SCHEMA_MAP = OrderedDict([
    (BooleanType, {"type": "boolean"}),
    (NumberType, {"type": "number"}),
    (BaseType, {"type": "string"}),
])


class SchematicsSerializer(ObjectSerializer):
    """
    An ObjectSerializer which allows the serialization of
    basic types and schematics models.

    The valid types that SchematicsSerializer supports are:

    - int
    - float
    - bool
    - decimal
    - string
    - none
    - lists, in the form of [Type] (e.g. [str])
    - any type that extends the schematics.models.Model.
    """

    def __init__(self, builtin_models=None):
        builtin_models = builtin_models or MODEL_MAP
        self._models = dict(builtin_models)

    def _translate_to_model(self, model):
        model = self._to_key(model)

        if isinstance(model, BaseType):
            return model

        if isinstance(model, ModelMeta):
            return ModelType(model)

        if model not in self._models and isinstance(model, tuple):
            self._models[model] = ListType(
                self._translate_to_model(model[0])
            )

        if model in self._models:
            return self._models[model]

        return model

    @staticmethod
    def _to_key(model):
        if isinstance(model, list):
            return tuple(model)
        return model

    def load(self, model, value):
        try:
            model = self._translate_to_model(model)
            result = model(value, context=get_import_context(
                oo=True
            ))
            if hasattr(result, "validate"):
                result.validate()
            return result
        except BaseError as e:
            raise SerializationException(str(e))

    def dump(self, model, value):
        try:
            model = self._translate_to_model(model)
            return model.to_primitive(value)
        except BaseError as e:
            raise SerializationException(str(e))

    def to_json_schema(self, model):
        model = self._translate_to_model(model)
        return _to_json_schema(model)

_cache = {}


def _to_json_schema(model):
    if model not in _cache:
        _cache[model] = _to_json_schema_no_cache(model)
    return _cache[model]


def _to_json_schema_no_cache(model):
    if isinstance(model, ModelType):
        return _model_type_to_json_schema(model)
    elif isinstance(model, ListType):
        return _list_type_to_json_schema(model)
    elif isinstance(model, DictType):
        return _dict_type_to_json_schema(model)
    else:
        for cls, schema in JSON_SCHEMA_MAP.items():
            if isinstance(model, cls):
                return schema


def _model_type_to_json_schema(model):
    required = []
    schema = {
        "title": model.model_name,
        "type": "object",
        "properties": {}
    }
    for name, field in model.fields.items():
        if field.required:
            required.append(name)
        schema["properties"][name] = _to_json_schema(field)
    if len(required) > 0:
        schema["required"] = required
    return schema


def _list_type_to_json_schema(list_type):
    return {
        "type": "array",
        "items": _to_json_schema(list_type.field)
    }


def _dict_type_to_json_schema(dict_type):
    return {
        "type": "object",
        "additionalProperties": _to_json_schema(dict_type.field)
    }



import os
import jinja2
from swagger_schema import Swagger, Info

CURDIR = os.path.dirname(__file__)


def generate_swagger_html(swagger_static_root, swagger_json_url):
    """
    given a root directory for the swagger statics, and
    a swagger json path, return back a swagger html designed
    to use those values.
    """
    tmpl = _get_template("swagger.html")
    return tmpl.render(
        swagger_root=swagger_static_root,
        swagger_json_url=swagger_json_url
    )


def get_swagger_static_root():
    """
    transmute-core includes the statics to render
    a swagger page. Use this function to
    return the directory containing said statics.
    """
    return os.path.join(CURDIR, "static")

_template_cache = {}


def _get_template(template_name):
    if template_name not in _template_cache:
        template_path = os.path.join(CURDIR, template_name)
        with open(template_path) as fh:
            _template_cache[template_name] = jinja2.Template(fh.read())
    return _template_cache[template_name]


class SwaggerSpec(object):
    """
    a class for aggregating and outputting swagger definitions, from
    transmute primitives
    """

    def __init__(self):
        self._swagger = {}

    def add_func(self, transmute_func, transmute_context):
        """ add a transmute function's swagger definition to the spec """
        swagger_path = transmute_func.get_swagger_path(transmute_context)
        for p in transmute_func.paths:
            if p not in self._swagger:
                self._swagger[p] = swagger_path
            else:
                for method, definition in swagger_path.items():
                    setattr(self._swagger[p], method, definition)

    @property
    def paths(self):
        """
        return the path section of the final swagger spec,
        aggregated from the paths added.
        """
        return self._swagger

    def swagger_definition(self, title="example", version="1.0"):
        """
        return a valid swagger spec, with the values passed.
        """
        return Swagger({
            "info": Info({"title": title, "version": version}),
            "paths": self.paths,
            "swagger": "2.0",
            "basePath": "/",
        }).to_primitive()



from transmute_core.function.parameters import Parameters
from transmute_core.function.signature import Argument, NoDefault
from transmute_core.swagger.function import get_swagger_parameters
from transmute_core import default_context


def test_swagger_parameters():
    parameters = Parameters(
        query={"query": Argument("query", None, int)},
        body={"body": Argument("body", NoDefault, int)},
        header={"header": Argument("header", NoDefault, int)},
        path={"path": Argument("path", NoDefault, int)},
    )
    params = get_swagger_parameters(parameters, default_context)
    params = [p.to_primitive() for p in params]
    assert {
        "in": "query",
        "name": "query",
        "required": False,
        "type": "number"
    } in params
    assert {
        "in": "body",
        "name": "body",
        "schema": {"type": "number"},
        "required": True,
    } in params
    assert {
        "in": "header",
        "name": "header",
        "type": "number",
        "required": True,
    } in params
    assert {
        "in": "path",
        "name": "path",
        "type": "number",
        "required": True,
    } in params
